#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import lxml.etree
import re
from unittest import mock

import requests_mock

import dracclient.client
from dracclient import constants
from dracclient import exceptions
from dracclient.resources import idrac_card
from dracclient.resources import job
from dracclient.resources import uris
from dracclient.tests import base
from dracclient.tests import utils as test_utils


@requests_mock.Mocker()
@mock.patch.object(dracclient.client.WSManClient,
                   'wait_until_idrac_is_ready', spec_set=True,
                   autospec=True)
class ClientiDRACCardConfigurationTestCase(base.BaseTest):

    def setUp(self):
        super(ClientiDRACCardConfigurationTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    def test_list_idrac_settings_by_instance_id(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_enum_attr = idrac_card.iDRACCardEnumerableAttribute(
            name='Type',
            instance_id='iDRAC.Embedded.1#Info.1#Type',
            read_only=True,
            current_value='13G Monolithic',
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='Info.1',
            possible_values=['12G/13G', '12G Monolithic', '12G Modular',
                             '13G Monolithic', '13G Modular', '12G DCS',
                             '13G DCS'])
        expected_string_attr = idrac_card.iDRACCardStringAttribute(
            name='Version',
            instance_id='iDRAC.Embedded.1#Info.1#Version',
            read_only=True,
            current_value='2.40.40.40',
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='Info.1',
            min_length=0,
            max_length=63)
        expected_integer_attr = idrac_card.iDRACCardIntegerAttribute(
            name='Port',
            instance_id='iDRAC.Embedded.1#SSH.1#Port',
            read_only=False,
            current_value=22,
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='SSH.1',
            lower_bound=1,
            upper_bound=65535)
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])

        idrac_settings = self.drac_client.list_idrac_settings()

        self.assertEqual(631, len(idrac_settings))
        # enumerable attribute
        self.assertIn('iDRAC.Embedded.1#Info.1#Type', idrac_settings)
        self.assertEqual(expected_enum_attr, idrac_settings[
                         'iDRAC.Embedded.1#Info.1#Type'])
        # string attribute
        self.assertIn('iDRAC.Embedded.1#Info.1#Version', idrac_settings)
        self.assertEqual(expected_string_attr,
                         idrac_settings['iDRAC.Embedded.1#Info.1#Version'])
        # integer attribute
        self.assertIn('iDRAC.Embedded.1#SSH.1#Port', idrac_settings)
        self.assertEqual(expected_integer_attr, idrac_settings[
                         'iDRAC.Embedded.1#SSH.1#Port'])

    def test_list_idrac_settings_by_name(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_enum_attr = idrac_card.iDRACCardEnumerableAttribute(
            name='Type',
            instance_id='iDRAC.Embedded.1#Info.1#Type',
            read_only=True,
            current_value='13G Monolithic',
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='Info.1',
            possible_values=['12G/13G', '12G Monolithic', '12G Modular',
                             '13G Monolithic', '13G Modular', '12G DCS',
                             '13G DCS'])
        expected_string_attr = idrac_card.iDRACCardStringAttribute(
            name='Version',
            instance_id='iDRAC.Embedded.1#Info.1#Version',
            read_only=True,
            current_value='2.40.40.40',
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='Info.1',
            min_length=0,
            max_length=63)
        expected_integer_attr = idrac_card.iDRACCardIntegerAttribute(
            name='Port',
            instance_id='iDRAC.Embedded.1#SSH.1#Port',
            read_only=False,
            current_value=22,
            pending_value=None,
            fqdd='iDRAC.Embedded.1',
            group_id='SSH.1',
            lower_bound=1,
            upper_bound=65535)
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])

        idrac_settings = self.drac_client.list_idrac_settings(by_name=True)

        self.assertEqual(630, len(idrac_settings))

        # enumerable attribute
        self.assertIn('Info.1#Type', idrac_settings)
        self.assertEqual(expected_enum_attr, idrac_settings[
                         'Info.1#Type'])
        # string attribute
        self.assertIn('Info.1#Version', idrac_settings)
        self.assertEqual(expected_string_attr,
                         idrac_settings['Info.1#Version'])
        # integer attribute
        self.assertIn('SSH.1#Port', idrac_settings)
        self.assertEqual(expected_integer_attr, idrac_settings[
                         'SSH.1#Port'])

    def test_list_multi_idrac_settings_by_name(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_enum_attr = idrac_card.iDRACCardEnumerableAttribute(
            name='Type',
            instance_id='iDRAC.Embedded.2#Info.1#Type',
            read_only=True,
            current_value='13G Monolithic',
            pending_value=None,
            fqdd='iDRAC.Embedded.2',
            group_id='Info.1',
            possible_values=['12G/13G', '12G Monolithic', '12G Modular',
                             '13G Monolithic', '13G Modular', '12G DCS',
                             '13G DCS'])
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])

        idrac_settings = self.drac_client.list_idrac_settings(
            by_name=True, fqdd_filter='iDRAC.Embedded.2')

        self.assertEqual(1, len(idrac_settings))

        # enumerable attribute
        self.assertIn('Info.1#Type', idrac_settings)
        self.assertEqual(expected_enum_attr, idrac_settings[
                         'Info.1#Type'])

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_set_idrac_settings(
            self, mock_requests, mock_invoke, mock_wait_until_idrac_is_ready):
        expected_selectors = {'CreationClassName': 'DCIM_iDRACCardService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:iDRACCardService',
                              'SystemCreationClassName': 'DCIM_ComputerSystem'}
        expected_properties = {'Target': 'iDRAC.Embedded.1',
                               'AttributeName': ['LDAP.1#GroupAttributeIsDN'],
                               'AttributeValue': ['Disabled']}
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.iDracCardInvocations[uris.DCIM_iDRACCardService][
                'SetAttributes']['ok'])

        result = self.drac_client.set_idrac_settings(
            {'LDAP.1#GroupAttributeIsDN': 'Disabled'})

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required':
                              constants.RebootRequired.false},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_iDRACCardService, 'SetAttributes',
            expected_selectors, expected_properties,
            wait_for_idrac=True)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke',
                       spec_set=True, autospec=True)
    def test_set_idrac_settings_with_valid_length_string(
            self, mock_requests, mock_invoke, mock_wait_until_idrac_is_ready):
        expected_selectors = {'CreationClassName': 'DCIM_iDRACCardService',
                              'SystemName': 'DCIM:ComputerSystem',
                              'Name': 'DCIM:iDRACCardService',
                              'SystemCreationClassName': 'DCIM_ComputerSystem'}
        expected_properties = {'Target': 'iDRAC.Embedded.1',
                               'AttributeName': ['Users.16#Password'],
                               'AttributeValue': ['12345678901234567890']}
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.iDracCardInvocations[uris.DCIM_iDRACCardService][
                'SetAttributes']['ok'])
        result = self.drac_client.set_idrac_settings(
            {'Users.16#Password': '12345678901234567890'})
        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required':
                              constants.RebootRequired.false},
                         result)
        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_iDRACCardService, 'SetAttributes',
            expected_selectors, expected_properties,
            wait_for_idrac=True)

    def test_set_idrac_settings_with_too_long_string(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = ("Attribute 'Password' cannot be set to "
                            "value '123456789012345678901'. It must be "
                            "between 0 and 20 characters in length.")
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardEnumeration]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardString]['ok']},
            {'text': test_utils.iDracCardEnumerations[
                uris.DCIM_iDRACCardInteger]['ok']}])
        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_idrac_settings,
            {'Users.16#Password': '123456789012345678901'})


class ClientiDRACCardChangesTestCase(base.BaseTest):

    def setUp(self):
        super(ClientiDRACCardChangesTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    @mock.patch.object(job.JobManagement, 'create_config_job', spec_set=True,
                       autospec=True)
    def test_commit_pending_idrac_changes(self, mock_create_config_job):
        self.drac_client.commit_pending_idrac_changes()

        mock_create_config_job.assert_called_once_with(
            mock.ANY,
            resource_uri=uris.DCIM_iDRACCardService,
            cim_creation_class_name='DCIM_iDRACCardService',
            cim_name='DCIM:iDRACCardService',
            target=dracclient.client.DRACClient.IDRAC_FQDD,
            reboot=False, start_time='TIME_NOW')

    @mock.patch.object(job.JobManagement, 'create_config_job', spec_set=True,
                       autospec=True)
    def test_commit_pending_idrac_changes_with_reboot(
            self, mock_create_config_job):

        self.drac_client.commit_pending_idrac_changes(
            reboot=True)

        mock_create_config_job.assert_called_once_with(
            mock.ANY,
            resource_uri=uris.DCIM_iDRACCardService,
            cim_creation_class_name='DCIM_iDRACCardService',
            cim_name='DCIM:iDRACCardService',
            target=dracclient.client.DRACClient.IDRAC_FQDD,
            reboot=True, start_time='TIME_NOW')

    @mock.patch.object(job.JobManagement, 'create_config_job', spec_set=True,
                       autospec=True)
    def test_commit_pending_idrac_changes_with_time(
            self, mock_create_config_job):
        timestamp = '20140924120101'
        self.drac_client.commit_pending_idrac_changes(
            start_time=timestamp)

        mock_create_config_job.assert_called_once_with(
            mock.ANY,
            resource_uri=uris.DCIM_iDRACCardService,
            cim_creation_class_name='DCIM_iDRACCardService',
            cim_name='DCIM:iDRACCardService',
            target=dracclient.client.DRACClient.IDRAC_FQDD,
            reboot=False, start_time=timestamp)

    @mock.patch.object(job.JobManagement, 'create_config_job', spec_set=True,
                       autospec=True)
    def test_commit_pending_idrac_changes_with_reboot_and_time(
            self, mock_create_config_job):

        timestamp = '20140924120101'
        self.drac_client.commit_pending_idrac_changes(
            reboot=True,
            start_time=timestamp)

        mock_create_config_job.assert_called_once_with(
            mock.ANY,
            resource_uri=uris.DCIM_iDRACCardService,
            cim_creation_class_name='DCIM_iDRACCardService',
            cim_name='DCIM:iDRACCardService',
            target=dracclient.client.DRACClient.IDRAC_FQDD,
            reboot=True, start_time=timestamp)

    @mock.patch.object(job.JobManagement, 'delete_pending_config',
                       spec_set=True, autospec=True)
    def test_abandon_pending_idrac_changes(self, mock_delete_pending_config):
        self.drac_client.abandon_pending_idrac_changes()

        mock_delete_pending_config.assert_called_once_with(
            mock.ANY,
            resource_uri=uris.DCIM_iDRACCardService,
            cim_creation_class_name='DCIM_iDRACCardService',
            cim_name='DCIM:iDRACCardService',
            target=dracclient.client.DRACClient.IDRAC_FQDD)


class ClientiDRACCardResetTestCase(base.BaseTest):

    def setUp(self):
        super(ClientiDRACCardResetTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    @mock.patch('dracclient.client.subprocess.call')
    def test_ping_host(self, mock_os_system):
        mock_os_system.return_value = 0
        response = self.drac_client._ping_host('127.0.0.1')
        self.assertEqual(mock_os_system.call_count, 1)
        self.assertEqual(True, response)

    @mock.patch('dracclient.client.subprocess.call')
    def test_ping_host_not_pingable(self, mock_os_system):
        mock_os_system.return_value = 1
        response = self.drac_client._ping_host('127.0.0.1')
        self.assertEqual(mock_os_system.call_count, 1)
        self.assertEqual(False, response)

    @mock.patch('dracclient.client.subprocess.call')
    def test_ping_host_name_not_known(self, mock_os_system):
        mock_os_system.return_value = 2
        response = self.drac_client._ping_host('127.0.0.1')
        self.assertEqual(mock_os_system.call_count, 1)
        self.assertEqual(False, response)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_alive(self, mock_ping_host, mock_sleep):
        total_calls = 5
        ping_count = 3
        mock_ping_host.return_value = True
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=True,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(True, response)
        self.assertEqual(mock_sleep.call_count, ping_count)
        self.assertEqual(mock_ping_host.call_count, ping_count)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_alive_fail(self, mock_ping_host, mock_sleep):
        total_calls = 5
        ping_count = 3
        mock_ping_host.return_value = False
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=True,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(False, response)
        self.assertEqual(mock_sleep.call_count, total_calls)
        self.assertEqual(mock_ping_host.call_count, total_calls)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_dead(self, mock_ping_host, mock_sleep):
        total_calls = 5
        ping_count = 3
        mock_ping_host.return_value = False
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=False,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(True, response)
        self.assertEqual(mock_sleep.call_count, ping_count)
        self.assertEqual(mock_ping_host.call_count, ping_count)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_dead_fail(self, mock_ping_host, mock_sleep):
        total_calls = 5
        ping_count = 3
        mock_ping_host.return_value = True
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=False,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(False, response)
        self.assertEqual(mock_sleep.call_count, total_calls)
        self.assertEqual(mock_ping_host.call_count, total_calls)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_alive_with_intermittent(
            self, mock_ping_host, mock_sleep):
        total_calls = 6
        ping_count = 3
        mock_ping_host.side_effect = [True, True, False, True, True, True]
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=True,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(True, response)
        self.assertEqual(mock_sleep.call_count, total_calls)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.DRACClient._ping_host')
    def test_wait_for_host_dead_with_intermittent(
            self, mock_ping_host, mock_sleep):
        total_calls = 6
        ping_count = 3
        mock_ping_host.side_effect = [False, False, True, False, False, False]
        mock_sleep.return_value = None
        response = self.drac_client._wait_for_host_state(
            'hostname',
            alive=False,
            ping_count=ping_count,
            retries=total_calls)
        self.assertEqual(True, response)
        self.assertEqual(mock_sleep.call_count, total_calls)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_reset_idrac(self, mock_invoke):
        expected_selectors = {
            'CreationClassName': "DCIM_iDRACCardService",
            'Name': "DCIM:iDRACCardService",
            'SystemCreationClassName': 'DCIM_ComputerSystem',
            'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Force': '0'}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.iDracCardInvocations[uris.DCIM_iDRACCardService][
                'iDRACReset']['ok'])

        result = self.drac_client.reset_idrac()

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_iDRACCardService, 'iDRACReset',
            expected_selectors, expected_properties,
            check_return_value=False)
        self.assertTrue(result)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_reset_idrac_force(self, mock_invoke):
        expected_selectors = {
            'CreationClassName': "DCIM_iDRACCardService",
            'Name': "DCIM:iDRACCardService",
            'SystemCreationClassName': 'DCIM_ComputerSystem',
            'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Force': '1'}
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.iDracCardInvocations[uris.DCIM_iDRACCardService][
                'iDRACReset']['ok'])

        result = self.drac_client.reset_idrac(force=True)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_iDRACCardService, 'iDRACReset',
            expected_selectors, expected_properties,
            check_return_value=False)
        self.assertTrue(result)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_reset_idrac_bad_result(self, mock_invoke):
        expected_selectors = {
            'CreationClassName': "DCIM_iDRACCardService",
            'Name': "DCIM:iDRACCardService",
            'SystemCreationClassName': 'DCIM_ComputerSystem',
            'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Force': '0'}
        expected_message = ("Failed to reset iDRAC")
        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.iDracCardInvocations[uris.DCIM_iDRACCardService][
                'iDRACReset']['error'])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.reset_idrac)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_iDRACCardService, 'iDRACReset',
            expected_selectors, expected_properties,
            check_return_value=False)

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.WSManClient.wait_until_idrac_is_ready')
    @mock.patch('dracclient.client.DRACClient._wait_for_host_state')
    @mock.patch(
        'dracclient.client.idrac_card.iDRACCardConfiguration.reset_idrac')
    def test_reset_idrac_wait(
            self,
            mock_reset_idrac,
            mock_wait_for_host_state,
            mock_wait_until_idrac_is_ready,
            mock_sleep):
        mock_reset_idrac.return_value = True
        mock_wait_for_host_state.side_effect = [True, True]
        mock_wait_until_idrac_is_ready.return_value = True
        mock_sleep.return_value = None

        self.drac_client.reset_idrac(wait=True)

        mock_reset_idrac.assert_called_once()
        self.assertEqual(mock_wait_for_host_state.call_count, 2)
        mock_wait_until_idrac_is_ready.assert_called_once()

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.WSManClient.wait_until_idrac_is_ready')
    @mock.patch('dracclient.client.DRACClient._wait_for_host_state')
    @mock.patch(
        'dracclient.client.idrac_card.iDRACCardConfiguration.reset_idrac')
    def test_reset_idrac_wait_failed_reset(
            self,
            mock_reset_idrac,
            mock_wait_for_host_state,
            mock_wait_until_idrac_is_ready,
            mock_sleep):
        mock_reset_idrac.return_value = False
        mock_wait_for_host_state.side_effect = [True, True]
        mock_wait_until_idrac_is_ready.return_value = False
        mock_sleep.return_value = None
        expected_message = ("Failed to reset iDRAC")

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.reset_idrac, wait=True)

        mock_reset_idrac.assert_called_once()
        mock_wait_for_host_state.assert_not_called()
        mock_wait_until_idrac_is_ready.assert_not_called()

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.WSManClient.wait_until_idrac_is_ready')
    @mock.patch('dracclient.client.DRACClient._wait_for_host_state')
    @mock.patch(
        'dracclient.client.idrac_card.iDRACCardConfiguration.reset_idrac')
    def test_reset_idrac_fail_wait_not_pingable(
            self,
            mock_reset_idrac,
            mock_wait_for_host_state,
            mock_wait_until_idrac_is_ready,
            mock_sleep):
        mock_reset_idrac.return_value = True
        mock_wait_for_host_state.side_effect = [False, True]
        mock_wait_until_idrac_is_ready.return_value = True
        mock_sleep.return_value = None
        expected_message = (
            "Timed out waiting for the 1.2.3.4 iDRAC to become not pingable")

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.reset_idrac, wait=True)

        mock_reset_idrac.assert_called_once()
        mock_wait_for_host_state.assert_called_once()
        mock_wait_until_idrac_is_ready.assert_not_called()

    @mock.patch('time.sleep')
    @mock.patch('dracclient.client.WSManClient.wait_until_idrac_is_ready')
    @mock.patch('dracclient.client.DRACClient._wait_for_host_state')
    @mock.patch(
        'dracclient.client.idrac_card.iDRACCardConfiguration.reset_idrac')
    def test_reset_idrac_fail_wait_pingable(
            self,
            mock_reset_idrac,
            mock_wait_for_host_state,
            mock_wait_until_idrac_is_ready,
            mock_sleep):
        mock_reset_idrac.return_value = True
        mock_wait_for_host_state.side_effect = [True, False]
        mock_wait_until_idrac_is_ready.return_value = True
        mock_sleep.return_value = None
        expected_message = (
            "Timed out waiting for the 1.2.3.4 iDRAC to become pingable")

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.reset_idrac, wait=True)

        mock_reset_idrac.assert_called_once()
        self.assertEqual(mock_wait_for_host_state.call_count, 2)
        mock_wait_until_idrac_is_ready.assert_not_called()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import lxml.etree
import re
from unittest import mock

import requests_mock

import dracclient.client
from dracclient import constants
from dracclient import exceptions
from dracclient.resources.inventory import NIC
import dracclient.resources.job
from dracclient.resources import nic
from dracclient.resources import uris
from dracclient.tests import base
from dracclient.tests import utils as test_utils


@requests_mock.Mocker()
@mock.patch.object(dracclient.client.WSManClient, 'wait_until_idrac_is_ready',
                   spec_set=True, autospec=True)
class ClientNICTestCase(base.BaseTest):

    def setUp(self):
        super(ClientNICTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    def test_list_nics(self, mock_requests,
                       mock_wait_until_idrac_is_ready):
        expected_nics = [
            NIC(id='NIC.Embedded.1-1-1',
                mac='B0:83:FE:C6:6F:A1',
                model='Broadcom Gigabit Ethernet BCM5720 - B0:83:FE:C6:6F:A1',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Slot.2-1-1',
                mac='A0:36:9F:52:7D:1E',
                model='Intel(R) Gigabit 2P I350-t Adapter - A0:36:9F:52:7D:1E',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Slot.2-2-1',
                mac='A0:36:9F:52:7D:1F',
                model='Intel(R) Gigabit 2P I350-t Adapter - A0:36:9F:52:7D:1F',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Embedded.2-1-1',
                mac='B0:83:FE:C6:6F:A2',
                model='Broadcom Gigabit Ethernet BCM5720 - B0:83:FE:C6:6F:A2',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T')
        ]

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.InventoryEnumerations[uris.DCIM_NICView]['ok'])

        self.assertEqual(expected_nics, self.drac_client.list_nics())

    def test_list_nics_sorted(self, mock_requests,
                              mock_wait_until_idrac_is_ready):
        expected_nics = [
            NIC(id='NIC.Embedded.1-1-1',
                mac='B0:83:FE:C6:6F:A1',
                model='Broadcom Gigabit Ethernet BCM5720 - B0:83:FE:C6:6F:A1',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Embedded.2-1-1',
                mac='B0:83:FE:C6:6F:A2',
                model='Broadcom Gigabit Ethernet BCM5720 - B0:83:FE:C6:6F:A2',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Slot.2-1-1',
                mac='A0:36:9F:52:7D:1E',
                model='Intel(R) Gigabit 2P I350-t Adapter - A0:36:9F:52:7D:1E',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T'),
            NIC(id='NIC.Slot.2-2-1',
                mac='A0:36:9F:52:7D:1F',
                model='Intel(R) Gigabit 2P I350-t Adapter - A0:36:9F:52:7D:1F',
                speed_mbps=1000,
                duplex='full duplex',
                media_type='Base T')
        ]

        mock_requests.post(
            'https://1.2.3.4:443/wsman',
            text=test_utils.InventoryEnumerations[uris.DCIM_NICView]['ok'])

        self.assertEqual(expected_nics, self.drac_client.list_nics(sort=True))

    def test_list_nic_settings(self, mock_requests,
                               mock_wait_until_idrac_is_ready):
        expected_enum_attr = nic.NICEnumerationAttribute(
            name='BootStrapType',
            instance_id='NIC.Integrated.1-3-1:BootStrapType',
            fqdd='NIC.Integrated.1-3-1',
            read_only=False,
            current_value='AutoDetect',
            pending_value=None,
            possible_values=['AutoDetect', 'BBS', 'Int18h', 'Int19h'])

        expected_string_attr = nic.NICStringAttribute(
            name='BusDeviceFunction',
            instance_id='NIC.Integrated.1-3-1:BusDeviceFunction',
            fqdd='NIC.Integrated.1-3-1',
            read_only=True,
            current_value='02:00:00',
            pending_value=None,
            min_length=0,
            max_length=0,
            pcre_regex=None)

        expected_integer_attr = nic.NICIntegerAttribute(
            name='BlnkLeds',
            instance_id='NIC.Integrated.1-3-1:BlnkLeds',
            fqdd='NIC.Integrated.1-3-1',
            read_only=False,
            current_value=0,
            pending_value=None,
            lower_bound=0,
            upper_bound=15)

        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        nic_settings = self.drac_client.list_nic_settings(
            nic_id='NIC.Integrated.1-3-1')

        self.assertEqual(63, len(nic_settings))

        self.assertIn('BootStrapType', nic_settings)
        self.assertEqual(expected_enum_attr,
                         nic_settings['BootStrapType'])

        self.assertIn('BusDeviceFunction', nic_settings)
        self.assertEqual(expected_string_attr,
                         nic_settings['BusDeviceFunction'])

        self.assertIn('BlnkLeds', nic_settings)
        self.assertEqual(expected_integer_attr,
                         nic_settings['BlnkLeds'])

    def test_list_nic_settings_with_colliding_attrs(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['colliding']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaises(
            exceptions.DRACOperationFailed,
            self.drac_client.list_nic_settings,
            nic_id='NIC.Integrated.1-3-1')

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_set_nic_settings(self, mock_requests, mock_invoke,
                              mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        expected_selectors = {'CreationClassName': 'DCIM_NICService',
                              'Name': 'DCIM:NICService',
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': 'NIC.Integrated.1-3-1',
                               'AttributeValue': ['PXE'],
                               'AttributeName': ['LegacyBootProto']}

        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.NICInvocations[uris.DCIM_NICService][
                'SetAttributes']['ok'])

        result = self.drac_client.set_nic_settings(
            nic_id='NIC.Integrated.1-3-1',
            settings={'LegacyBootProto': 'PXE'})

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_NICService, 'SetAttributes',
            expected_selectors, expected_properties,
            wait_for_idrac=True)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_set_nic_settings_string(self, mock_requests, mock_invoke,
                                     mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        expected_selectors = {'CreationClassName': 'DCIM_NICService',
                              'Name': 'DCIM:NICService',
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': 'NIC.Integrated.1-3-1',
                               'AttributeValue': ['D4:AE:52:A5:B1:01'],
                               'AttributeName': ['VirtMacAddr']}

        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.NICInvocations[uris.DCIM_NICService][
                'SetAttributes']['ok'])

        result = self.drac_client.set_nic_settings(
            nic_id='NIC.Integrated.1-3-1',
            settings={'VirtMacAddr': 'D4:AE:52:A5:B1:01'})

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_NICService, 'SetAttributes',
            expected_selectors, expected_properties,
            wait_for_idrac=True)

    @mock.patch.object(dracclient.client.WSManClient, 'invoke', spec_set=True,
                       autospec=True)
    def test_set_nic_settings_integer(self, mock_requests, mock_invoke,
                                      mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        expected_selectors = {'CreationClassName': 'DCIM_NICService',
                              'Name': 'DCIM:NICService',
                              'SystemCreationClassName': 'DCIM_ComputerSystem',
                              'SystemName': 'DCIM:ComputerSystem'}
        expected_properties = {'Target': 'NIC.Integrated.1-3-1',
                               'AttributeValue': [1],
                               'AttributeName': ['SecondTgtBootLun']}

        mock_invoke.return_value = lxml.etree.fromstring(
            test_utils.NICInvocations[uris.DCIM_NICService][
                'SetAttributes']['ok'])

        result = self.drac_client.set_nic_settings(
            nic_id='NIC.Integrated.1-3-1',
            settings={'SecondTgtBootLun': 1})

        self.assertEqual({'is_commit_required': True,
                          'is_reboot_required': constants.RebootRequired.true},
                         result)

        mock_invoke.assert_called_once_with(
            mock.ANY, uris.DCIM_NICService, 'SetAttributes',
            expected_selectors, expected_properties,
            wait_for_idrac=True)

    def test_set_nic_settings_error(self, mock_requests,
                                    mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']},
            {'text': test_utils.NICInvocations[
                uris.DCIM_NICService]['SetAttributes']['error']}])

        self.assertRaises(exceptions.DRACOperationFailed,
                          self.drac_client.set_nic_settings,
                          'NIC.InvalidTarget',
                          {'LegacyBootProto': 'PXE'})

    def test_set_nic_settings_with_unknown_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaises(exceptions.InvalidParameterValue,
                          self.drac_client.set_nic_settings,
                          'NIC.Integrated.1-3-1',
                          {'foo': 'bar'})

    def test_set_nic_settings_with_unchanged_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        result = self.drac_client.set_nic_settings(
            nic_id='NIC.Integrated.1-3-1',
            settings={'LegacyBootProto': 'NONE'})

        self.assertEqual({'is_commit_required': False,
                          'is_reboot_required':
                          constants.RebootRequired.false},
                         result)

    def test_set_nic_settings_with_readonly_attr(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = (
            "Cannot set read-only iDRAC Card attributes: ['LinkStatus'].")
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_nic_settings,
            'NIC.Integrated.1-3-1',
            {'LinkStatus': 'Connected'})

    def test_set_nic_settings_with_incorrect_enum(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = (
            "Attribute 'LegacyBootProto' cannot be set to value 'foo'. "
            "It must be in ['PXE', 'iSCSI', 'NONE'].")
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_nic_settings,
            'NIC.Integrated.1-3-1',
            {'LegacyBootProto': 'foo'})

    def test_set_nic_settings_with_incorrect_regexp(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = (
            "Attribute 'VirtMacAddr' cannot be set to value 'foo.' "
            "It must match regex '^([0-9a-fA-F]{2}:){5}([0-9a-fA-F]{2})$'.")
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_nic_settings,
            'NIC.Integrated.1-3-1',
            {'VirtMacAddr': 'foo'})

    def test_set_nic_settings_with_out_of_bounds_value(
            self, mock_requests, mock_wait_until_idrac_is_ready):
        expected_message = (
            "Attribute BannerMessageTimeout cannot be set to value 100. "
            "It must be between 0 and 14.")
        mock_requests.post('https://1.2.3.4:443/wsman', [
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICEnumeration]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICString]['ok']},
            {'text': test_utils.NICEnumerations[
                uris.DCIM_NICInteger]['ok']}])

        self.assertRaisesRegexp(
            exceptions.DRACOperationFailed, re.escape(expected_message),
            self.drac_client.set_nic_settings,
            'NIC.Integrated.1-3-1',
            {'BannerMessageTimeout': 100})


class ClientNICSettingTestCase(base.BaseTest):

    def setUp(self):
        super(ClientNICSettingTestCase, self).setUp()
        self.drac_client = dracclient.client.DRACClient(
            **test_utils.FAKE_ENDPOINT)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_create_nic_config_job(self, mock_create_config_job):

        nic_id = 'NIC.Embedded.1-1-1'
        self.drac_client.create_nic_config_job(
            nic_id=nic_id)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_NICService,
            cim_creation_class_name='DCIM_NICService',
            cim_name='DCIM:NICService',
            target=nic_id,
            reboot=False,
            start_time='TIME_NOW')

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_create_nic_config_job_reboot(self, mock_create_config_job):

        nic_id = 'NIC.Embedded.1-1-1'
        self.drac_client.create_nic_config_job(
            nic_id=nic_id,
            reboot=True)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_NICService,
            cim_creation_class_name='DCIM_NICService',
            cim_name='DCIM:NICService',
            target=nic_id,
            reboot=True,
            start_time='TIME_NOW')

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_create_nic_config_job_time(self, mock_create_config_job):

        nic_id = 'NIC.Embedded.1-1-1'
        timestamp = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
        self.drac_client.create_nic_config_job(
            nic_id=nic_id,
            start_time=timestamp)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_NICService,
            cim_creation_class_name='DCIM_NICService',
            cim_name='DCIM:NICService',
            target=nic_id,
            reboot=False,
            start_time=timestamp)

    @mock.patch.object(dracclient.resources.job.JobManagement,
                       'create_config_job', spec_set=True, autospec=True)
    def test_create_nic_config_job_no_time(self, mock_create_config_job):

        nic_id = 'NIC.Embedded.1-1-1'
        self.drac_client.create_nic_config_job(
            nic_id=nic_id,
            start_time=None)

        mock_create_config_job.assert_called_once_with(
            mock.ANY, resource_uri=uris.DCIM_NICService,
            cim_creation_class_name='DCIM_NICService',
            cim_name='DCIM:NICService',
            target=nic_id,
            reboot=False,
            start_time=None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class BaseClientException(Exception):

    msg_fmt = 'An unknown exception occurred'

    def __init__(self, message=None, **kwargs):
        message = self.msg_fmt % kwargs
        super(BaseClientException, self).__init__(message)


class DRACRequestFailed(BaseClientException):
    pass


class DRACOperationFailed(DRACRequestFailed):
    msg_fmt = ('DRAC operation failed. Messages: %(drac_messages)s')


class DRACUnexpectedReturnValue(DRACRequestFailed):
    msg_fmt = ('DRAC operation yielded return value %(actual_return_value)s '
               'that is neither error nor the expected '
               '%(expected_return_value)s')


class DRACEmptyResponseField(BaseClientException):
    msg_fmt = ("Attribute '%(attr)s' is not nullable, but no value received")


class DRACMissingResponseField(BaseClientException, AttributeError):
    msg_fmt = ("Attribute '%(attr)s' is missing from the response")


class InvalidParameterValue(BaseClientException):
    msg_fmt = '%(reason)s'


class WSManRequestFailure(BaseClientException):
    msg_fmt = ('WSMan request failed')


class WSManInvalidResponse(BaseClientException):
    msg_fmt = ('Invalid response received. Status code: "%(status_code)s", '
               'reason: "%(reason)s"')


class WSManInvalidFilterDialect(BaseClientException):
    msg_fmt = ('Invalid filter dialect "%(invalid_filter)s". '
               'Supported options are %(supported)s')
